package graphoutput;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.io.*;
import javax.media.j3d.*;
import javax.vecmath.*;
import com.sun.j3d.utils.geometry.*;

 /**
  * Branchgroup which contains an even plane fitted into the 1x1x1-box. The
  * orientation and translation of the plane is freely chooseable.
  */
class CuttingPlane extends BranchGroup
{
  /**
   * Use white as default color
   */
  private final static Color3f DEFAULT_COL = new Color3f(1.0f, 1.0f, 1.0f);
  /**
   * Set the default transparency
   */
  private final static float  DEFAULT_TRANSPARENCY = 0.2f;
  /**
   * Set the epsilon value,
   * used for comparison if 2 points are on the same location
   */
  private final static double EPSILON = 0.01;

  private Point3d    p;
  private Vector3d   v;
  private Appearance app, app1;
  private Shape3D    plane, plane1;
   /**
    * Contains a textual description of this plane
    */
  public  String     description;

  /**
   * Constructor for default-values
   */
  public CuttingPlane ()
  {
    p = new Point3d(0.5, 0.5, 0.5);
    v = new Vector3d(0.0, 0.0, 1.0);
    v.normalize();
    description = new String("p=("+p.x+", "+p.y+", "+p.z+")  n=("+v.x+", "+v.y+", "+v.z+")");

    app = createAppearance(DEFAULT_COL, DEFAULT_TRANSPARENCY);
    app1 = createAppearanceBS(DEFAULT_COL, DEFAULT_TRANSPARENCY);
    createPlane();
    setCapability(BranchGroup.ALLOW_DETACH);
    setCapability(BranchGroup.ALLOW_CHILDREN_READ);
    setCapability(BranchGroup.ALLOW_CHILDREN_WRITE);
    setCapability(BranchGroup.ALLOW_CHILDREN_EXTEND);
    compile();
  }

  /**
   *  Constructor where the position of the cutting plane is
   *  specified.
   */
  public CuttingPlane(Point3d p, Vector3d v)
  {
    this.p=p;
    this.v=v;
    generateAutoDescription();

    app = createAppearance(DEFAULT_COL, DEFAULT_TRANSPARENCY);
    app1 = createAppearanceBS(DEFAULT_COL, DEFAULT_TRANSPARENCY);
    createPlane();
    setCapability(BranchGroup.ALLOW_DETACH);
    setCapability(BranchGroup.ALLOW_CHILDREN_READ);
    setCapability(BranchGroup.ALLOW_CHILDREN_WRITE);
    setCapability(BranchGroup.ALLOW_CHILDREN_EXTEND);
    compile();
  }

  /**
   *  Constructor where the position of the cutting plane
   *  and the color attributes are specified.
   */
  public CuttingPlane(Point3d p, Vector3d v, Color3f col, float transparency)
  {
    this.p=p;
    this.v=v;
    generateAutoDescription();

    app = createAppearance(col, transparency);
    app1 = createAppearanceBS(DEFAULT_COL, DEFAULT_TRANSPARENCY);
    createPlane();
    setCapability(BranchGroup.ALLOW_DETACH);
  }

  /**
   *  Constructor where the position of the cutting plane,
   *  the color attributes and the description string are specified.
   */
  public CuttingPlane(Point3d p, Vector3d v, Color3f col, float transparency, String description)
  {
    this.p=p;
    this.v=v;

    app = createAppearance(col, transparency);
    app1 = createAppearanceBS(DEFAULT_COL, DEFAULT_TRANSPARENCY);
    createPlane();
    setCapability(BranchGroup.ALLOW_DETACH);
  }

   /**
    * Generates a text containing a description of the plane
    */
  public void generateAutoDescription()
  {
    description = new String("p=("+p.x+", "+p.y+", "+p.z+")  n=("+v.x+", "+v.y+", "+v.z+")");
  }

  /**
   *  Sets the description string of the cutting-plane
   */
  public void setDescription(String description)
  {
    this.description = description;
  }

  /**
   *  Set the position of the CuttingPlane
   */
  public void setPosition(Point3d p, Vector3d v)
  {
    this.p=p;
    this.v=v;
    createPlane();
  }

  /**
   *  Returns the current point of the cutting-plane
   */
  public Point3d getPoint()
  {
    return p;
  }

  /**
   *  Returns the current normal vector of the cutting-plane
   */
  public Vector3d getNormalVector()
  {
    return v;
  }

  /**
   *  Set the color and the transparency of the CuttingPlane
   */
  public void setColor(Color3f col, float transparency)
  {
    app = createAppearance(col, transparency);
    app1 = createAppearanceBS(col, transparency);
    createPlane();
  }

  /**
   * Creates an instance of Appearance with the specified
   * attributes
   */
  private Appearance createAppearance(Color3f color, float transparency)
  {
    Appearance app = new Appearance();

    Material mat = new Material(color,
                                new Color3f(0.0f,0.0f,0.0f),
                                color,
                                new Color3f(1.0f,1.0f,1.0f),
                                80.0f);
    mat.setCapability(Material.ALLOW_COMPONENT_READ);
    app.setMaterial(mat);
    app.setCapability(Appearance.ALLOW_MATERIAL_READ);
    app.setCapability(Appearance.ALLOW_TRANSPARENCY_ATTRIBUTES_READ);
    PolygonAttributes pa = new PolygonAttributes();

    TransparencyAttributes ta = new TransparencyAttributes();
    ta.setTransparencyMode(ta.BLENDED);
    ta.setTransparency(transparency);
    ta.setCapability(TransparencyAttributes.ALLOW_VALUE_READ);
    app.setTransparencyAttributes(ta);

    pa.setCullFace(pa.CULL_FRONT);
    app.setPolygonAttributes(pa);

    return app;
  }

  /**
   * Creates an instance of Appearance with the specified
   * attributes for the backside
   */
  private Appearance createAppearanceBS(Color3f color, float transparency)
  {
    Appearance app = new Appearance();

    Material mat = new Material(color,
                                new Color3f(0.0f,0.0f,0.0f),
                                color,
                                new Color3f(1.0f,1.0f,1.0f),
                                80.0f);
    mat.setCapability(Material.ALLOW_COMPONENT_READ);
    app.setMaterial(mat);
    app.setCapability(Appearance.ALLOW_MATERIAL_READ);
    app.setCapability(Appearance.ALLOW_TRANSPARENCY_ATTRIBUTES_READ);
    PolygonAttributes pa = new PolygonAttributes();

    TransparencyAttributes ta = new TransparencyAttributes();
    ta.setTransparencyMode(ta.BLENDED);
    ta.setTransparency(transparency);
    ta.setCapability(TransparencyAttributes.ALLOW_VALUE_READ);
    app.setTransparencyAttributes(ta);

    pa.setCullFace(pa.CULL_BACK);
    app.setPolygonAttributes(pa);

    return app;
  }

  /**
   * Returns the color of the cuttingplane
   */
  public Color3f getColor()
  {
    Color3f col = new Color3f();
    Material mat = app.getMaterial();
    mat.getDiffuseColor(col);
    return col;
  }

  /**
   * Returns the transparency of the cuttingplane
   */
  public float getTransparency()
  {
    TransparencyAttributes ta = app.getTransparencyAttributes();
    return ta.getTransparency();
  }


  /**
   * Creates the Shape3D of the cutting-plane and adds it to
   * this BranchGroup. This method uses a heuristic method to
   * obtain the intersection-points of a plane and the 1x1x1-box.
   */
  private void createPlane()
  {
    // needed while calculating
    double x,y,z;

    // Vector which will contain all intersection-points
    Vector is_pts = new Vector();
    // ordered Vector with all necessary intersection-points
    Vector od_pts = new Vector();

    // intersect plane with all 12 edges of the 1x1x1-cube
    double d = v.x*p.x + v.y*p.y + v.z*p.z;

    // 1
    if(v.y!=0)
    { y = d / v.y;
      if (y>=0 && y<=1) is_pts.addElement(new Point3d(0.0, y, 0.0)); }
    // 2
    if(v.x!=0)
    { x = (d-v.y)/v.x;
      if (x>=0 && x<=1) is_pts.addElement(new Point3d(x, 1.0, 0.0)); }
    // 3
    if(v.y!=0)
    { y = (d-v.x)/v.y;
      if (y>=0 && y<=1) is_pts.addElement(new Point3d(1.0, y, 0.0)); }
    // 4
    if(v.x!=0)
    { x = d / v.x;
      if (x>=0 && x<=1) is_pts.addElement(new Point3d(x, 0.0, 0.0)); }
    // 5
    if(v.z!=0)
    { z = (d-v.y)/ v.z;
      if (z>=0 && z<=1) is_pts.addElement(new Point3d(0.0, 1.0, z)); }
    // 6
    if(v.z!=0)
    { z = (d-v.x-v.y)/ v.z;
      if (z>=0 && z<=1) is_pts.addElement(new Point3d(1.0, 1.0, z)); }
    // 7
    if(v.z!=0)
    { z = (d-v.x)/ v.z;
      if (z>=0 && z<=1) is_pts.addElement(new Point3d(1.0, 0.0, z)); }
    // 8
    if(v.z!=0)
    { z = d / v.z;
      if (z>=0 && z<=1) is_pts.addElement(new Point3d(0.0, 0.0, z)); }
    // 9
    if(v.y!=0)
    { y = (d-v.z)/ v.y;
      if (y>=0 && y<=1) is_pts.addElement(new Point3d(0.0, y, 1.0)); }
    // 10
    if(v.x!=0)
    { x = (d-v.y-v.z)/ v.x;
      if (x>=0 && x<=1) is_pts.addElement(new Point3d(x, 1.0, 1.0)); }
    // 11
    if(v.y!=0)
    { y = (d-v.x-v.z)/ v.y;
      if (y>=0 && y<=1) is_pts.addElement(new Point3d(1.0, y, 1.0)); }
    // 12
    if(v.x!=0)
    { x = (d-v.z)/ v.x;
      if (x>=0 && x<=1) is_pts.addElement(new Point3d(x, 0.0, 1.0)); }

    // remove all duplicates
    for(int i=0; i<is_pts.size(); i++)
    {
      Point3d p1 = (Point3d) is_pts.elementAt(i);
      for (int j=i+1; j<is_pts.size(); j++)
      {
        Point3d p2 = (Point3d) is_pts.elementAt(j);
        if(p1.distanceSquared(p2)<(EPSILON*EPSILON))
          is_pts.removeElementAt(j);
      }
    }

    // now create an ordered list for drawing
    // a polygon

    od_pts.addElement((Point3d)is_pts.elementAt(0));
    Point3d cur_pt =(Point3d)is_pts.elementAt(0);

    is_pts.removeElementAt(0);

    while(is_pts.size()>0)
    {
      double mindist=10.0;
      int    min_ndx=0;

      for(int j=0;j<is_pts.size(); j++)
      {
        Point3d p = (Point3d)is_pts.elementAt(j);
        if(cur_pt.distanceSquared(p) < mindist)
        {
          mindist = cur_pt.distanceSquared(p);
          min_ndx = j;
        }
      }
      od_pts.addElement((Point3d)is_pts.elementAt(min_ndx));
      cur_pt = (Point3d)is_pts.elementAt(min_ndx);
      is_pts.removeElementAt(min_ndx);
    }

    int[] q;
    q = new int[1];
    q[0]=od_pts.size();

    TriangleFanArray tfa = new TriangleFanArray(od_pts.size(), GeometryArray.COORDINATES | GeometryArray.NORMALS, q);
    TriangleFanArray tfa1 = new TriangleFanArray(od_pts.size(), GeometryArray.COORDINATES | GeometryArray.NORMALS, q);

	for(int i=0; i<od_pts.size(); i++)
    {
     tfa.setCoordinate(i, (Point3d)od_pts.elementAt(i) );
     tfa1.setCoordinate(i, (Point3d)od_pts.elementAt(i) );
    }

    Point3f k = new Point3f((float)((Point3d)od_pts.elementAt(0)).x, (float)((Point3d)od_pts.elementAt(0)).y, (float)((Point3d)od_pts.elementAt(0)).z);
    Point3f l = new Point3f((float)((Point3d)od_pts.elementAt(1)).x, (float)((Point3d)od_pts.elementAt(1)).y, (float)((Point3d)od_pts.elementAt(1)).z);
    Point3f m = new Point3f((float)((Point3d)od_pts.elementAt(2)).x, (float)((Point3d)od_pts.elementAt(2)).y, (float)((Point3d)od_pts.elementAt(2)).z);

    Vector3f v1 = new Vector3f();
    Vector3f v2 = new Vector3f();
    Vector3f n = new Vector3f();

    v1.sub(l, k);
    v2.sub(m, k);
    n.cross(v1, v2);
    n.normalize();

    for (int i=0; i<od_pts.size()-2; i++)
    {
      tfa.setNormal(i, n);
      tfa1.setNormal(i, n);
    }

    plane = new Shape3D();
    plane.setGeometry(tfa);
    plane.setAppearance(app);
    plane1 = new Shape3D();
    plane1.setGeometry(tfa1);
    plane1.setAppearance(app1);

    BranchGroup bg1 = new BranchGroup();
    bg1.setCapability(BranchGroup.ALLOW_DETACH);
    bg1.addChild(plane);
    bg1.addChild(plane1);

    if(numChildren()!=0)
	{
	  for (int i=numChildren()-1; i>=0; i--)
       ((BranchGroup)this.getChild(i)).detach();
    }


    addChild(bg1);
  }
} // End of class CuttingPlane